package dk.brics.xact;

import dk.brics.misc.Origin;

/**
 * Base class for all nodes.
 */
public abstract class Node { // TODO: need equals and hashCode?
	
	private final Origin origin;
	
	private volatile Node replacement; // always maps an XML node to an XML node and an AttrNode node to an AttrNode node
	
	private volatile boolean replaced;
	
	// TODO: add "has_been_normalized" flag
	
	/**
	 * Constructs a new node.
	 */
	public Node(Origin origin) {
		this.origin = origin;
	}
	
	/**
	 * Returns a description of this node.
	 */
	@Override
	abstract public String toString();
	
	/**
	 * Returns the replacement node if forwarded, otherwise returns the node itself.
	 * All methods in this package should invoke <code>follow</code> before inspecting the type of a node.
	 * Public methods in this package should never return a forwarded node.
	 */
	static Node follow(Node n) {
		while (n != null && n.replaced)
			n = n.replacement;
		return n;
	}

	/**
	 * Returns the replacement node if forwarded, otherwise returns the node itself.
	 * See {@link #follow(Node)}.
	 */
	static XML follow(XML x) {
		return (XML)follow((Node)x);
	}

	/**
	 * Returns the replacement node if forwarded, otherwise returns the node itself.
	 * See {@link #follow(Node)}.
	 */
	static AttrNode follow(AttrNode a) {
		return (AttrNode)follow((Node)a);
	}

	/**
	 * Returns the replacement node if forwarded, otherwise returns the node itself.
	 * See {@link #follow(Node)}.
	 */
	static NamespaceDecl follow(NamespaceDecl a) {
		return (NamespaceDecl)follow((Node)a);
	}

	/**
	 * Returns the origin of this node, or null if absent.
	 */
	public Origin getOrigin() {
		return origin;
	}

	/**
	 * Forwards this node.
	 * This is the only way to modify a node!
	 * Should only be invoked on a freshly generated (which implies non-shared) node.
	 */
	void forward(Node replacement) {
		this.replacement = replacement;
		replaced = true;
	}
	
	/**
	 * Checks whether this is an element node.
	 */
	public boolean isElement() {
		return follow(this) instanceof Element;
	}
	
	/**
	 * Returns this node as an element node.
	 * @throws ClassCastException if this is not an element node
	 */
	public Element asElement() {
		return (Element)follow(this);
	}
	
	/**
	 * Checks whether this is a text node.
	 */
	public boolean isText() {
		return follow(this) instanceof Text;
	}
	
	/**
	 * Returns this node as a text node.
	 * @throws ClassCastException if this is not a text node
	 */
	public Text asText() {
		return (Text)follow(this);
	}
	
	/**
	 * Checks whether this is an comment node.
	 */
	public boolean isComment() {
		return follow(this) instanceof Comment;
	}
	
	/**
	 * Returns this node as a comment node.
	 * @throws ClassCastException if this is not a comment node
	 */
	public Comment asComment() {
		return (Comment)follow(this);
	}
	
	/**
	 * Checks whether this is an processing instruction node.
	 */
	public boolean isProcessingInstruction() {
		return follow(this) instanceof ProcessingInstruction;
	}
	
	/**
	 * Returns this node as a processing instruction node.
	 * @throws ClassCastException if this is not a processing instruction node
	 */
	public ProcessingInstruction asProcessingInstruction() {
		return (ProcessingInstruction)follow(this);
	}
	
	/**
	 * Checks whether this is an attribute node.
	 */
	public boolean isAttribute() {
		return follow(this) instanceof Attribute;
	}
	
	/**
	 * Returns this node as an attribute node.
	 * @throws ClassCastException if this is not an attribute node
	 */
	public Attribute asAttribute() {
		return (Attribute)follow(this);
	}
	
	/**
	 * Checks whether this is a namespace declaration node.
	 */
	public boolean isNamespaceDecl() {
		return follow(this) instanceof NamespaceDecl;
	}
	
	/**
	 * Returns this node as a namespace declaration node.
	 * @throws ClassCastException if this is not a namespace declaration node
	 */
	public NamespaceDecl asNamespaceDecl() {
		return (NamespaceDecl)follow(this);
	}
	
	/**
	 * Checks whether this is an XML template content node.
	 */
	boolean isTemplate() {
		return follow(this) instanceof TempNode;
	}
	
	/**
	 * Returns this node as an XML template content node.
	 * @throws ClassCastException if this is not an XML node
	 */
	TempNode asTemplate() {
		return (TempNode)follow(this);
	}
	
	/**
	 * Visits this node by the given visitor.
	 */
	public void visitBy(NodeVisitor v) {
		follow(this).visitNormalizedBy(v);
	}
	/**
	 * Visits this node by the given visitor.
	 */
	public <R,A> R visitBy(NodeVisitorParameterized<R,A> v, A arg) {
		return follow(this).visitNormalizedBy(v, arg);
	}
	
	/**
	 * Visits this normalized node it by the given visitor.
	 */
	abstract void visitNormalizedBy(NodeVisitor v);
	
	/**
	 * Visits this normalized node it by the given visitor.
	 */
	abstract <R,A> R visitNormalizedBy(NodeVisitorParameterized<R,A> v, A arg);

	/**
	 * Visits this node by the given any-node visitor.
	 */
	abstract void visitAnyBy(AnyNodeVisitor v);

	/**
	 * Builds an expanded name from an optional namespace URI and a local name.
	 */
	static String expanded(String namespace, String localname) {
		if (namespace == null || namespace.length() == 0)
			return localname;
		else
			return "{" + namespace + "}" + localname;
	}

	/**
	 * @deprecated for debugging only
	 */
	@Deprecated
	public void dump(int indent) { // XXX: remove this method, for debugging only
		Node n = follow(this);
		StringBuilder ind = new StringBuilder();
		for (int i = 0; i < indent; i++)
			ind.append(' ');
		System.out.println(ind.toString() + n);
		if (n instanceof Element) {
			Element e = (Element)n;
			if (e.getFirstNamespaceDecl() != null)
				e.getFirstNamespaceDecl().dump(indent + 1);
			if (e.getFirstAttr() != null) 
				e.getFirstAttr().dump(indent + 1);
			if (e.getRealFirstChild() != null)
				e.getRealFirstChild().dump(indent + 1);
		}
		if (n instanceof PlugSingleNode) {
			PlugSingleNode p = (PlugSingleNode)n;
			if (p.getValue() instanceof XML) {
				System.out.println(ind.toString() + " [");
				((XML)p.getValue()).dump(indent + 2);
				System.out.println(ind.toString() + " ]");
			}
			if (p.getSelfNode() != null)
				p.getSelfNode().dump(indent + 1);
		}
		if (n instanceof PlugListNode) {
			PlugListNode p = (PlugListNode)n;
			System.out.println(ind.toString() + " [");
			for (Object v : p.getValues())
				if (v instanceof XML) {
					((XML)v).dump(indent + 2);
					System.out.println(ind.toString() + "  ;");
				}
			System.out.println(ind.toString() + " ]");
			if (p.getSelfNode() != null)
				p.getSelfNode().dump(indent + 1);
		}
		if (n instanceof CloseNode) {
			CloseNode p = (CloseNode)n;
			if (p.getSelfNode() != null)
				p.getSelfNode().dump(indent + 1);
		}
		if (n instanceof ConcatNode) {
			ConcatNode c = (ConcatNode)n;
			c.getLeftNode().dump(indent + 1);
			c.getRightNode().dump(indent + 1);
		}
		if (n instanceof TempNode) {
			TempNode t = (TempNode)n;
			if (t.getRealNextSibling() != null)
				t.getRealNextSibling().dump(indent);
		} else if (n instanceof AttrNode) {
			AttrNode a = (AttrNode)n;
			if (a.getNextAttr() != null)
				a.getNextAttr().dump(indent);
		}
	}
	
} 
