package dk.brics.xact;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Stack;

/**
 * Removes operation nodes.
 */
class Normalizer {

	/**
	 * Special object used to mark gap removals.
	 */
	private static final Object PLUG_REMOVE = new Object();

	/**
	 * Special object used to mark gaps that remain open.
	 */
	private static final Object PLUG_OPEN = new Object();

	/**
	 * Next-sibling link. Used when building results bottom-up.
	 */
	private XML nextsibling;

	/**
	 * Next-attr link. Used when building results bottom-up.
	 */
	private AttrNode nextattr;

	/**
	 * Heap stack, avoids deep recursive calls.
	 */
	private Stack<Entry> stack; 

	/**
	 * Children for each ancestor.
	 */
	private Stack<XML> childstack; 

	/**
	 * Attributes for each ancestor.
	 */
	private Stack<AttrNode> attrstack;

	/**
	 * Map from gap name (null means 'other') to stack of ancestor plugs/closes.
	 */
	private Map<String, Stack<PlugInfo>> plugs; 

	/**
	 * Values for gaps. Pushed on the way down, popped on the way back.
	 */
	private Stack<Object> valuestack; 

	/**
	 * Additional following siblings for each ancestor. Used when removing concat nodes.
	 */
	private Stack<Stack<XML>> moresiblings;

	private Normalizer() {}

	static XML normalize(XML x) {
		Normalizer n = new Normalizer();
		return n.removeConcats(n.removePlugs(x));
	}

	// TODO: skip traversal if we can detect that no (relevant) gaps are reachable
	/**
	 * Removes plug-like nodes (but preserves concat nodes).
	 */
	private XML removePlugs(XML x) {
		nextsibling = null;
		nextattr = null;
		stack = new Stack<Entry>();
		childstack = new Stack<XML>();
		attrstack = new Stack<AttrNode>();
		plugs = new HashMap<String, Stack<PlugInfo>>();
		valuestack = new Stack<Object>();

		// start traversal
		pushEntry(Kind.START_NODE, x);
		while (!stack.isEmpty()) {
			final Entry e = stack.pop();
			Node n = Node.follow(e.node);
			switch (e.kind) {

			case START_NODE: // enter a node

				// make sure we also visit on the way back
				pushEntry(Kind.END_NODE, n);

				//  when done with this node, visit the next attr/sibling
				n.visitAnyBy(new AnyNodeVisitor() {
					
					@Override
					public void visit(AttrNode n) {
						pushEntry(Kind.START_NODE, n.getNextAttr());
					}

					@Override
					public void visit(TempNode n) {
						pushEntry(Kind.START_NODE, n.getRealNextSibling());
					}
				});

				n.visitAnyBy(new AnyNodeVisitor() {

					// for a ConcatNode, visit left node as content and then proceed with right node
					@Override
					public void visit(ConcatNode n) {
						pushEntry(Kind.START_NODE, n.getRightNode());
						pushEntry(Kind.END_CONTENT, n);
						pushEntry(Kind.START_NODE, n.getLeftNode());
						pushEntry(Kind.START_CONTENT, n);
					}

					// for a CloseNode, visit content
					@Override
					public void visit(CloseNode n) {
						pushEntry(Kind.END_CONTENT, n);
						pushEntry(Kind.START_NODE, n.getSelfNode());
						pushEntry(Kind.START_CONTENT, n);
					}

					// for a PlugSingleNode, visit the value (if XML) and then the contents
					@Override
					public void visit(PlugSingleNode n) {
						pushEntry(Kind.END_CONTENT, n);
						pushEntry(Kind.START_NODE, n.getSelfNode());
						PlugInfo pi = new PlugSingleInfo();
						pushEntry(Kind.START_CONTENT, n, pi);
						pushEntry(Kind.END_VALUES, n);
						Object v = n.getValue();
						pushEntry(Kind.END_VALUE, n, pi, v);
						if (v instanceof XML) {
							pushEntry(Kind.START_NODE, ((XML) v));
						}
						pushEntry(Kind.START_VALUES, n);
					}

					// for a PlugListNode, visit each value (if XML) and then the contents
					@Override
					public void visit(PlugListNode n) {
						pushEntry(Kind.END_CONTENT, n);
						pushEntry(Kind.START_NODE, n.getSelfNode());
						PlugInfo pi = new PlugListInfo();
						pushEntry(Kind.START_CONTENT, n, pi);
						pushEntry(Kind.END_VALUES, n);
						for (Object v : n.getValues()) {
							pushEntry(Kind.END_VALUE, n, pi, v);
							if (v instanceof XML) {
								pushEntry(Kind.START_NODE, ((XML) v));
							}
						}
						pushEntry(Kind.START_VALUES, n);
					}

					// for an Element, visit attributes and then contents
					@Override
					public void visit(Element n) {
						pushEntry(Kind.END_CONTENT, n);
						pushEntry(Kind.START_NODE, n.getRealFirstChild());
						pushEntry(Kind.START_CONTENT, n);
						pushEntry(Kind.START_NODE, n.getFirstAttr());
					}

					// for an AttributeGap, pick a value and push it onto the valuestack
					@Override
					public void visit(AttributeGap n) {
						pushGapValue(n.getGap());
					}

					// for a TemplateGap, pick a value and push it onto the valuestack
					@Override
					public void visit(TemplateGap n) {
						pushGapValue(n.getGap());
					}
				});
				break;

			case START_VALUES: // enter plug values	
				
				// start a new sibling sequence
				childstack.push(nextsibling);
				nextsibling = null;
				break;

			case END_VALUE: // a plug value is completed
				
				// if the value is XML, grab it from the sibling sequence, otherwise from the original value
				Object v;
				if (e.v instanceof XML) {
					v = nextsibling;
					nextsibling = null;
				} else {
					v = e.v;
				}
				e.pi.addValue(v);
				break;

			case END_VALUES: // leave plug values

				// return to previous sibling sequence
				nextsibling = childstack.pop();
				break;

			case START_CONTENT: // enter contents of an element/plug/concat
				
				n.visitAnyBy(new AnyNodeVisitor() {
					
					// for a PlugNode, push pluginfo onto plug stack of the gap
					@Override
					public void visit(PlugNode n) {
						pushPlug(n.getGap(), e.pi);
					}

					// for a PlugNode, push close instruction onto plug stack of the gap
					@Override
					public void visit(CloseNode n) {
						pushClose();
					}

					// for an Element, attrs are now completed so push current attr sequence onto the attrstack
					@Override
					public void visit(Element n) {
						attrstack.push(nextattr);
						nextattr = null;
					}
				});
				break;

			case END_CONTENT: // leave contents of an element/plug/concat
				
				n.visitAnyBy(new AnyNodeVisitor() {
					
					// for a ConcatNode, left part is now completed so push current sibling sequence onto the childstack
					@Override
					public void visit(ConcatNode n) {
						childstack.push(nextsibling);
						nextsibling = null;
					}

					@Override
					public void visit(PlugNode n) {
						popPlug(n.getGap());
					}

					@Override
					public void visit(CloseNode n) {
						popClose();
					}

					// for an Element, contents are now completed so push current sibling sequence onto the childstack
					@Override
					public void visit(Element n) {
						childstack.push(nextsibling);
						nextsibling = null;
					}
				});
				break;

			case END_NODE: // leave a node
				
				n.visitAnyBy(new AnyNodeVisitor() {
					
					// for a ConcatNode, copy and insert into the current sibling sequence
					@Override
					public void visit(ConcatNode n) {
						nextsibling = n.copy(childstack.pop(), nextsibling);
					}

					// for an Attribute, copy and insert into the current attr sequence
					@Override
					public void visit(Attribute n) {
						nextattr = n.copy(nextattr);
					}

					// for an Element, copy and insert into the current sibling sequence
					@Override
					public void visit(Element n) {
						nextsibling = n.copy(attrstack.pop(), childstack.pop(), nextsibling);
					}

					// for an AttributeGap, take a value from valuestack and insert it into the current attr sequence unless open or removed
					@Override
					public void visit(AttributeGap n) {
						Object v = valuestack.pop();
						if (v == PLUG_OPEN) {
							nextattr = n.copy(nextattr);
						} else if (v != PLUG_REMOVE) {
							if (v instanceof Text) {
								v = ((Text) v).getString();
							}
							if (v instanceof XML) {
								throw new XMLPlugException("plugging XML template into attribute gap '" + n.getGap() + "'", n.getOrigin());
							} 
							nextattr = new Attribute(n.getNamespace(), n.getLocalName(), (String) v, nextattr, n.getOrigin());
						}
					}

					// for a TemplateGap, take a value from valuestack and insert it into the current sibling sequence unless open or removed
					@Override
					public void visit(TemplateGap n) {
						Object v = valuestack.pop();
						if (v == PLUG_OPEN) {
							nextsibling = n.copy(nextsibling);
						} else {
							if (v == PLUG_REMOVE) {
								v = "";
							}
							if (v instanceof XML) {
								nextsibling = XML.concatXML((XML) v, nextsibling);
							} else {
								nextsibling = new Text((String) v, nextsibling, n.getOrigin());
							}
						}
					}

					// for a Text, copy and insert into the current sibling sequence
					@Override
					public void visit(Text n) {
						nextsibling = n.copy(nextsibling);
					}

					// for a Comment, copy and insert into the current sibling sequence
					@Override
					public void visit(Comment n) {
						nextsibling = n.copy(nextsibling);
					}

					// for a ProcessingInstruction, copy and insert into the current sibling sequence
					@Override
					public void visit(ProcessingInstruction n) {
						nextsibling = n.copy(nextsibling);
					}
				});
				break;
			}
		}
		if (nextsibling == null) {
			throw new XMLException("internal error, null result");
		}
		return nextsibling;
	}

	/**
	 * Pushes an entry onto the execution stack.
	 */
	private void pushEntry(Kind kind, Node n, PlugInfo pi, Object v) {
		if (n != null) {
			stack.push(new Entry(kind, n, pi, v));
		}
	}

	/**
	 * Pushes an entry (with no value) onto the execution stack.
	 */
	private void pushEntry(Kind kind, Node n, PlugInfo pi) {
		pushEntry(kind, n, pi, null);
	}

	/**
	 * Pushes an entry (with no plug info or value) onto the execution stack.
	 */
	private void pushEntry(Kind kind, Node n) {
		pushEntry(kind, n, null, null);
	}

	/**
	 * Pushes a pluginfo onto the plug stack of the given gap.
	 */
	private void pushPlug(String gap, PlugInfo pi) {
		Stack<PlugInfo> ps = plugs.get(gap);
		if (ps == null) {
			ps = new Stack<PlugInfo>();
			plugs.put(gap, ps);
		}
		ps.push(pi);
	}

	/**
	 * Pops the top of the plug stack of the given gap.
	 */
	private void popPlug(String gap) {
		plugs.get(gap).pop();
		if (plugs.get(gap).isEmpty()) {
			plugs.remove(gap);
		}
	}

	/**
	 * Pushes a close instruction (null) onto all plug stacks, including the null gap name.
	 */
	private void pushClose() {
		if (!plugs.containsKey(null)) {
			plugs.put(null, new Stack<PlugInfo>());
		}
		for (Stack<PlugInfo> sp : plugs.values()) {
			sp.push(null);
		}
	}

	/**
	 * Pops the top of all plug stacks.
	 */
	private void popClose() {
		for (Stack<PlugInfo> sp : plugs.values()) {
			sp.pop();
		}
		if (plugs.get(null).isEmpty()) {
			plugs.remove(null);
		}
	}

	/**
	 * Takes a plug value for the given gap name and pushes it onto valuestack.
	 */
	 private void pushGapValue(String gap) {
		 Object v;
		 Stack<PlugInfo> sp = plugs.get(gap);
		 if (sp != null && !sp.isEmpty()) {
			 PlugInfo pi = sp.peek();
			 if (pi == null) {
				 v = PLUG_REMOVE;
			 } else {
				 v = pi.getNext();
			 }
		 } else if (plugs.containsKey(null)) {
			 v = PLUG_REMOVE;
		 } else
			 v = PLUG_OPEN;
		 valuestack.push(v);
	 }

	 // TODO: skip traversal if we can detect that no concats are reachable and moresiblings is empty
	 /**
	  * Removes concat nodes. Assumes that there are no other operation nodes.
	  */
	 private XML removeConcats(XML x) {
		 nextsibling = null;
		 stack = new Stack<Entry>();
		 moresiblings = new Stack<Stack<XML>>();
		 moresiblings.push(new Stack<XML>());
		 childstack = new Stack<XML>();
		 
		 // start traversal
		 pushEntry(Kind.START_NODE, x);
		 while (!stack.isEmpty()) {
			 Entry e = stack.pop();
			 Node n = Node.follow(e.node);
			 switch (e.kind) {
			 
			 case START_NODE: // enter a node

				 n.visitAnyBy(new AnyNodeVisitor() {
					 
					 // before leaving a template node, visit the additional siblings and then the ordinary siblings
					 @Override
					 public void visit(TempNode n) {
						 pushEntry(Kind.END_NODE, n);
						 if (n.getRealNextSibling() != null) {
							 pushEntry(Kind.START_NODE, n.getRealNextSibling());
						 } else if (!moresiblings.peek().isEmpty()) {
							 pushEntry(Kind.START_NODE, moresiblings.peek().pop());
						 }
					 }
				 });
				 
				 n.visitAnyBy(new AnyNodeVisitor() {
					 
					 // for a ConcatNode, visit the left node and add push right node an additional sibling 
					 @Override
					 public void visit(ConcatNode n) {
						 moresiblings.peek().push(n.getRightNode());
						 pushEntry(Kind.START_NODE, n.getLeftNode());
					 }

					 // for an Element, visit the contents 
					 @Override
					 public void visit(Element n) {
						 pushEntry(Kind.END_CONTENT, n);
						 pushEntry(Kind.START_NODE, n.getRealFirstChild());
						 pushEntry(Kind.START_CONTENT, n, null);
					 }
				 });
				 break;
				 
			 case START_CONTENT: // enter contents of an element
				 
				 // push a new stack for additional siblings
				 moresiblings.push(new Stack<XML>());
				 nextsibling = null;
				 break;
				 
			 case END_CONTENT: // leave contents of an element
				 
				 // push sibling sequence as a child sequence
				 moresiblings.pop();
				 childstack.push(nextsibling);
				 nextsibling = null;
				 break;
				 
			 case END_NODE: // leave a node
				 
				 // copy the node and insert it into the current sibling sequence
				 n.visitAnyBy(new AnyNodeVisitor() {
					 
					 @Override
					 public void visit(Element n) {
						 nextsibling = n.copy(n.getFirstAttr(), childstack.pop(), nextsibling);
					 }

					 @Override
					 public void visit(Text n) {
						 nextsibling = n.copy(nextsibling);
					 }

					 @Override
					 public void visit(Comment n) {
						 nextsibling = n.copy(nextsibling);
					 }

					 @Override
					 public void visit(ProcessingInstruction n) {
						 nextsibling = n.copy(nextsibling);
					 }

					 @Override
					 public void visit(TemplateGap n) {
						 nextsibling = n.copy(nextsibling);
					 }
				 });
				 break;
			 default:
				 // not used
				 break;
			 }
		 }
		 if (nextsibling == null) {
			 throw new XMLException("internal error, null result");
		 }
		 return nextsibling;
	 }

	 private static enum Kind {
		 START_NODE, END_NODE, START_CONTENT, END_CONTENT, START_VALUES, END_VALUES, END_VALUE
	 };

	 private static class Entry {

		 private Kind kind;

		 private Node node;

		 private PlugInfo pi;

		 private Object v;

		 Entry(Kind kind, Node node, PlugInfo pi, Object v) {
			 this.kind = kind;
			 this.node = node;
			 this.pi = pi;
			 this.v = v;
		 }
	 }

	 private abstract static class PlugInfo {

		 abstract Object getNext();

		 abstract void addValue(Object v);
	 }

	 private static class PlugSingleInfo extends PlugInfo {

		 private Object value;

		 @Override
		 Object getNext() {
			 return value;
		 }

		 @Override
		 void addValue(Object v) {
			 value = v;
		 }
	 }

	 private static class PlugListInfo extends PlugInfo {

		 private LinkedList<Object> values;

		 private Iterator<?> it;

		 PlugListInfo() {
			 values = new LinkedList<Object>();
		 }

		 @Override
		 Object getNext() {
			 Object v;
			 if (it == null) {
				 it = values.iterator();
			 }
			 if (it.hasNext()) {
				 v = it.next();
			 } else {
				 v = "";
			 }
			 return v;
		 }

		 @Override
		 void addValue(Object v) {
			 values.addFirst(v);
		 }
	 }
}
