package dk.brics.xact.analysis;

import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.StringTokenizer;

import dk.brics.misc.Origin;

/**
 * Accepts error reports from the analysis. The errors are collected in a list that can
 * be accessed with {@link #getErrors()}. This class must be thread-safe. 
 */
public class ErrorHandler {
    private List<ErrorReport> errors = Collections.synchronizedList(new ArrayList<ErrorReport>());
    
    public void error(Origin origin, ErrorType type, String message, Object ... args) {
        String msg = args.length > 0 ? String.format(message, args) : message;
        errors.add(new ErrorReport(origin, type, msg));
    }
    public void error(Throwable t) {
        errors.add(new ErrorReport(t));
    }
    
    public List<ErrorReport> getErrors() {
        return errors;
    }
    
    public void printMessages(PrintStream out) {
        printMessages(new PrintWriter(out));
    }
    public void printMessages(PrintWriter out) {
        if (errors.isEmpty()) {
            out.println("No errors detected!");
            out.flush();
            return;
        }
        Collections.sort(errors, new ErrorReport.OriginComparator());
        String prevfile = "?";
        boolean first=true;
        for (ErrorReport error : errors) {
            if (!prevfile.equals(error.getOrigin().getFile())) {
                if (!first) {
                    out.println(" ");
                }
                prevfile = error.getOrigin().getFile();
                if (error.hasOrigin()) {
                    out.println(error.getOrigin().getFile());
                } else {
                    out.println("Other messages:");
                }
            }
            first=false;
            String prefix;
            if (error.hasOrigin()) {
                prefix = "Line " + error.getOrigin().getLine() + "> ";
            } else {
                prefix = "";
            }
            prefix = repeat(' ', 12 - prefix.length()) + prefix; // indent a bit
            out.print(prefix);
            int linenr=0;
            for (StringTokenizer tok = new StringTokenizer(error.getMessage(),"\r\n"); tok.hasMoreTokens(); ) {
                String line = tok.nextToken();
                if (linenr > 0) {
                    out.print("            ");
                }
                out.println(line);
                linenr++;
            }
        }
        out.println();
        out.println("Total number of error messages: " + errors.size());
        out.flush();
    }
    
    private String repeat(char c, int n) {
        StringBuilder b = new StringBuilder();
        for (int i=0; i<n; i++) {
            b.append(c);
        }
        return b.toString();
    }
    
}
