package dk.brics.xact.analysis;

import dk.brics.xact.XML;

public enum ErrorType {
	
	/**
	 * An XPath expression always returned an empty result.
	 */
	EMPTY_RESULT	("Empty result", true), // XXX exactly what are the other cases? e.g. "no non-whitespace text..."
	
	/**
	 * An XPath boolean test is always false.
	 */
	TEST_FAILS		("Test fails", true),
	
	/**
	 * {@link XML#toDocument()} was called on something that might not be well-formed
	 * because it did not have a unique root element. 
	 */
	NOT_UNIQUE_ROOT		("Well-formedness error", false),
	
	/**
	 * A gap might be plugged with XML data.
	 */
	XML_IN_ATTRIBUTE	("Well-formedness error", false),
	
	/**
	 * A plug operation had no effect because the specified gap is absent.
	 */
	MISSING_GAP		("Operation has no effect", true),
	
	/**
	 * A plug or close operation might violate the type restriction on a gap
	 */
	INVALID_PLUG_TYPE ("Plug operation can violate a gap type", false),
	
	/**
	 * An operation like <tt>plug</tt> was used without using the result for anything,
	 * which is a typical bug pattern.
	 * <pre>
	 * x.plug(..); // reports warning
	 * x = x.plug(..); // this is probably the programmer meant
	 * </pre>
	 */
	DISCARDED_RESULT ("Operation's result is discarded without being used", true),
	
	/**
	 * An XML fragment might not satisfy the XML type it was expected to.
	 */
	INVALID ("Validation error", false),
	
	/**
	 * An XML constant in the program was malformed.
	 */
	MALFORMED_XML_CONSTANT ("Malformed XML constant", true),
	
	/**
	 * An operation from the runtime system that the analysis can not yet handle
	 * was used.
	 */
	UNSUPPORTED_OPERATION ("An unsupported operation was used", true),
	
	OTHER ("Error", true)
	
	;
	
	private String message;
	
	private boolean definite;
	
	private ErrorType(String message, boolean definite) {
		this.message = message;
		this.definite = definite;
	}
	
	/**
	 * Returns true if an error of this type denotes something that will <i>definitely</i> be
	 * the case at runtime. In other words, definite errors should only be reported if the analysis
	 * finds that there definitely is a problem, while non-definite errors should be reported
	 * if the analysis simply fails to prove that there never is a problem.
	 * <p/>
	 * For example, {@link #EMPTY_RESULT} is definite because the result
	 * will definitely be empty every time at runtime. An error like {@link #INVALID} is not
	 * definite because the XML may not be invalid every time during runtime, or may not even
	 * be invalid at all, but was reported because the analysis failed to prove otherwise.
	 * <p/>
	 * Talk of soundness usually only makes sense for the non-definite error types, but informally
	 * one can say that it is wrong to report a definite error if there is not actually such an error.
	 * @return true if this error type is definite; false otherwise
	 */
	public boolean isDefinite() {
		return definite;
	}
	
	/**
	 * Returns a short human-readable description of this error type. 
	 * @return a short human-readable string
	 */
	public String getMessage() {
		return message;
	}
}
