package dk.brics.xact.analysis.config;

import java.net.URL;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;

import soot.SootClass;
import soot.SootField;
import soot.SootMethod;
import soot.ValueBox;
import soot.jimple.InvokeExpr;
import dk.brics.automaton.Automaton;
import dk.brics.xact.analysis.XMLAnalysis;

/**
 * The customizable settings of the analysis. This interface controls the assumptions made by
 * the analysis and affects precision, soundness, and efficiency.
 * <p/>
 * It is generally assumed that an instance of {@link XMLAnalysis} has its own {@link Configuration}
 * instance that is not shared with other analyses.
 * <p/>
 * All methods can be called concurrently by threads spawned by the {@link ExecutorService} 
 * assigned with {@link XMLAnalysis#setExecutorService(ExecutorService)} in addition to the thread that started
 * the analysis with {@link XMLAnalysis#analyze()}. A safe way around this is always
 * to declare all methods <tt>synchronized</tt> (though probably not the optimal solution).
 * 
 * @see StandardConfiguration
 */
public interface Configuration {
	/** 
	 * Returns true if the specified invocation may lead to a non-application class
	 * @param expr an invocation expression 
	 * @return false for precision, true for soundness
	 */
	boolean canCallExtern(InvokeExpr expr);
	
	/**
	 * Returns true if the specified method can be called from the VM or from a non-application class.
	 * <p/>
	 * Note that this should take class initializers (<tt>&lt;clinit&gt;</tt>) into account! Class
	 * initializers are generally always externally callable.
	 * @param method an application method
	 * @return false for precision, true for soundness
	 */
	boolean isExternallyCallable(SootMethod method);
	
	/**
	 * Returns true if the specified class or interface may have implementations
	 * in non-application code.
	 * <p/>
	 * Note that types like <tt>java.lang.Object</tt> are also queried with this method,
	 * so the implementation should not be too optimistic.
	 * @param clazz a class or interface
	 * @return false for precision, true for soundness
	 */
	boolean hasExternalSubtypes(SootClass clazz);
	
	/**
	 * Returns true if the specified non-abstract class denotes a collection that can be expected to
	 * behave as a <i>normal</i> collection. A normal collection in this sense is a collection that
	 * only contains elements added with the <tt>add</tt> and <tt>addAll</tt> methods from the 
	 * {@link Collection} and {@link List} interfaces and those added in a non-empty constructor. 
	 * Additionally, all its methods must adhere to their specifications in {@link Collection} and {@link List}.
	 * @param qualifiedName fully qualified name of a non-abstract class implementing {@link Collection} . Eg. <tt>java.util.ArrayList</tt>.
	 * @return true for precision, false for soundness
	 */
	boolean isNormalCollection(String qualifiedName);
	
	/**
	 * Returns the type of XML returned by the specified method, or null
	 * if any XML may be returned.
	 * @param method a non-application or application method returning a subtype of XML.
	 * @param annotation the type placed with a @Type annotation, or <tt>null</tt> if no @Type annotation is present
	 * @return a type string (expanded or using a namespace prefix) or null if no type should be used 
	 */
	String getMethodReturnType(SootMethod method, String annotation);
	
	/**
	 * Returns the type of XML required by the specified method parameter, or
	 * null if any XML may be given as parameter.
	 * @param method a non-application or application method
	 * @param param index of a parameter whose type is a subtype of XML
	 * @param annotation the type placed with a @Type annotation, or <tt>null</tt> if no @Type annotation is present
	 * @return a type string (expanded or using a namespace prefix) or null if no type should be used
	 */
	String getMethodParameterType(SootMethod method, int param, String annotation);
	
	/**
	 * Returns the type of XML allowed in the specified field.
	 * @param field a field
	 * @param annotation the type placed with a @Type annotation, or <tt>null</tt> if no @Type annotation is present
	 * @return a type string (expanded or using a namespace prefix) or null if no type should be used
	 */
	String getFieldType(SootField field, String annotation);
	
	/**
	 * Lets the configuration modify the namespace prefix map. The given map will initially
	 * contain all the prefixes the analysis found declared in the application classes.
	 * @param namespaces map from namespace prefixes to namespaces. May be modified.
	 */
	void modifyNamespaces(Map<String,String> namespaces);
	
	/**
	 * Returns the locations of some additional schemas that should be accesible to the application
	 * classes. Null may be returned instead of an empty collection.
	 * @return a collection with URLs to schemas, or <tt>null</tt>
	 */
	Collection<URL> getAdditionalSchemas();
	
	/**
	 * Returns the strings that may be returned by the specified type's <tt>toString</tt>
	 * method, or <tt>null</tt> if nothing special is known about the <tt>toString</tt> method.
	 * @param clazz a class
	 * @return an automaton or <tt>null</tt>
	 */
	Automaton resolveToString(SootClass clazz);
	
	/**
	 * Models the behaviour of <tt>toXML</tt> methods declared in non-application classes.
	 * <p/>
	 * The resulting string should be the name of a type using a namespace preferably declared with {@link #modifyNamespaces(Map)}
	 * and defined in a schema referred to by {@link #getAdditionalSchemas()}.
	 * <p/>
	 * This method should return <tt>null</tt> if nothing is known about the method.
	 * 
	 * @param toXmlMethod a <tt>toXML</tt> method declared in a non-application class
	 * @return <tt>null</tt> if nothing is known about the method, or a type string denoting the schema type of the result
	 */
	String resolveToXML(SootMethod toXmlMethod);
	
	/**
	 * Translates a schema location to another URL string (possibly the same).
	 * <p/>
	 * The analyzed program would often be executed from a different base directory
	 * than the one used to run the analysis; so to find the correct schema location
	 * the analysis needs to know what base directory to search in.
	 * @param location
	 * @return an URL string, not <tt>null</tt>
	 */
	String translateSchemaLocation(String location);
	
	/**
	 * Expressions in the analyzed program whose XML graph should be available
	 * so they can be analyzed outside the XACT analysis.
	 * @return set of soot expression
	 */
	Set<ValueBox> getHotspots();
}
