package dk.brics.xact.analysis.flowgraph;

import java.util.*;

/** 
 * Variable label set for flow graph edges. 
 */
public class VariableFilter {

	/** If true, variables are blacklisted, otherwise they are whitelisted */
	private boolean reversed;
	
	private Set<Variable> variables = new LinkedHashSet<Variable>();
	
	private int call_id; // only used for toString
	
	private Kind kind;
	
	/**
	 * Variable filter kinds.
	 */
	public enum Kind {

		/**
		 * Normal intra-procedural flow.
		 */
		NORMAL,
		
		/**
		 * Flow from method entry node to first ordinary node of the method body.
		 */
		ENTRY,
		
		/**
		 * Flow from method invocation to method entry node.
		 */
		CALL,
		
		/**
		 * Flow from method exit node to successor of method invocation node.
		 */
		RETURN,
	};
	
	/** 
	 * Constructs a normal variable label set, initially containing all variables.
	 */
	public VariableFilter() {
		this(true, Kind.NORMAL, 0);
	}
	
	/** 
	 * Constructs a normal variable label set.
	 * @param reversed indicates whether the set should initially contain
	 *                 all variables (<code>true</code>) or no variables (<code>false</code>).
	 */
	public VariableFilter(boolean reversed) {
		this(reversed, Kind.NORMAL, 0);
	}
	
	/** 
	 * Constructs a normal variable label set, initially containing a single variable.
	 */
	public VariableFilter(Variable v) {
		this(false, Kind.NORMAL, 0);
		variables.add(v);
	}
	/** 
	 * Constructs a normal variable label set, initially containing a single variable.
	 */
	public VariableFilter(Variable v, Kind kind) {
		this(false, kind, 0);
		variables.add(v);
	}
	
	/** 
	 * Constructs a variable label set with a specific kind.
	 * @param reversed indicates whether the set should initially contain
	 *                 all variables (<code>true</code>) or no variables (<code>false</code>).
	 * @param kind the kind of edge to make
	 */
	public VariableFilter(boolean reversed, Kind kind) {
		this(reversed, kind, 0);
	}
	
	/** 
	 * Constructs a variable label set with a specific kind and call id.
	 * @param reversed indicates whether the set should initially contain
	 *                 all variables (<code>true</code>) or no variables (<code>false</code>).
	 * @param kind the kind of edge to make
	 * @param id the call ID of the edge
	 */
	public VariableFilter(boolean reversed, Kind kind, int id) {
		this.reversed = reversed;
		this.kind = kind;
		this.call_id = id;
	}
	
	/** 
	 * Adds a variable to the label set.
	 * Ignored for reversed filters.
	 */
	public void addVariable(Variable var) {
		if (!reversed) 
			variables.add(var);
	}
	
	/** 
	 * Removes a variable from the label set.
	 * Ignored for non-reversed filters.
	 */
	public void removeVariable(Variable var) {
		if (reversed)
			variables.add(var);
	}
	
	/** 
	 * Checks whether the given variable is contained in the label set.
	 */
	public boolean containsVariable(Variable var) {
		if (reversed) 
			return !variables.contains(var);
		else 
			return variables.contains(var);
	}
	
	/**
	 * Returns true if this definitely contains all variables. A non-reversed variable filter
	 * always returns false regardless of the number of variables added to it.
	 */
	public boolean containsAll() {
		return reversed && variables.size() == 0;
	}
	
	public boolean isReversed() {
		return reversed;
	}
	
	/**
	 * For reversed filters, returns the set of variables not in the filter, and for non-reversed filters,
	 * returns the set of variables in the filter.
	 */
	public Set<Variable> getVariableSet() {
		return variables;
	}
	
	/** 
	 * Returns the edge kind.
	 */
	public Kind getKind() {
		return kind;
	}
	
	/**
	 * Creates a new variable filter accepting only variables accepted
	 * by both variable filters. The input variable filters are unmodified.
	 * The resulting variable filter's kind will <i>always</i> be set to NORMAL.
	 * @param other a variable filter. Will not be modified
	 * @return a new variable filter instance
	 */
	public VariableFilter intersection(VariableFilter other) {
		VariableFilter vf = new VariableFilter();
		if (reversed && other.reversed) {
			vf.reversed = true;
			vf.variables.addAll(this.variables);
			vf.variables.addAll(other.variables);
		}
		else if (reversed && !other.reversed) {
			vf.reversed = false;
			vf.variables.addAll(other.variables);
			vf.variables.removeAll(this.variables);
		}
		else if (!reversed && other.reversed) {
			vf.reversed = false;
			vf.variables.addAll(this.variables);
			vf.variables.removeAll(other.variables);
		}
		else if (!reversed && !other.reversed) {
			vf.reversed = false;
			vf.variables.addAll(this.variables);
			vf.variables.retainAll(other.variables);
		}
		return vf;
	}
	
	/**
	 * Checks whether the given variable filter is equal to this variable filter.
	 */
	@Override
	public boolean equals(Object other) {
		if (other instanceof VariableFilter) {
			VariableFilter o = (VariableFilter)other;
			return reversed == o.reversed
			&& variables.equals(o.variables)
			&& kind.equals(o.kind)
			&& call_id == o.call_id;
		} else
			return false;
	}
	
	/**
	 * Returns a hash code for this variable filter.
	 */
	@Override
	public int hashCode() {
		return variables.hashCode()+kind.hashCode()+call_id*11;
	}
	
	/**
	 * Returns a string description of this variable filter.
	 */
	@Override
	public String toString() {
		StringBuilder b = new StringBuilder();
		switch (kind) {
		case NORMAL:
			break;
		case ENTRY:
			b.append("entry");
			break;
		case CALL:
			b.append("call#").append(call_id);
			break;
		case RETURN:
			b.append("return#").append(call_id);
			break;
		}
		if (reversed) 
			b.append('^');
		b.append('{');
		boolean first = true;
		for (Variable v : variables) {
			if (first)
				first = false;
			else
				b.append(',');
			b.append(v);
		}
		b.append('}');
		return b.toString();
	}
}
