package dk.brics.xact.analysis.xmlgraph;

/**
 * Represents the lattice:
 * <pre>
 *      MAYBE
 *      /   \
 *    YES    NO
 *      \   /
 *      BOTTOM
 * </pre>
 * The meaning of "yes" and "no" is not specified here &ndash; it depends on
 * the context of its use.
 */
public enum EBooleanLattice {
	MAYBE, // NOTE: the order of declaration is significant! we are using ordinal() in the tables below.
	YES,
	NO,
	BOTTOM;
	
	private static final EBooleanLattice[][] LeastUpperBound = {
		{MAYBE, MAYBE, MAYBE, MAYBE},
		{MAYBE, YES,   MAYBE, YES},
		{MAYBE, MAYBE, NO,    NO},
		{MAYBE, YES,   NO,    BOTTOM}
	};
	private static final EBooleanLattice[][] GreatestLowerBound = {
		{MAYBE, YES,   NO, 	  BOTTOM},
		{YES,   YES,   BOTTOM,BOTTOM},
		{NO,    BOTTOM,NO,    BOTTOM},
		{BOTTOM,BOTTOM,BOTTOM,BOTTOM}
	};
	private static final EBooleanLattice[][] And = {
		{MAYBE, MAYBE, NO,    BOTTOM},
		{MAYBE, YES,   NO,    BOTTOM},
		{NO,    NO,    NO,    BOTTOM},
		{BOTTOM,BOTTOM,BOTTOM,BOTTOM}
	};
	private static final EBooleanLattice[][] Or = {
		{MAYBE, YES,   MAYBE, BOTTOM},
		{YES,   YES,   YES,   BOTTOM},
		{MAYBE, YES,   NO,    BOTTOM},
		{BOTTOM,BOTTOM,BOTTOM,BOTTOM}
	};
	
	/**
	 * Returns the least upper bound of two lattice points.
	 */
	public EBooleanLattice leastUpperBound(EBooleanLattice lat) {
		return LeastUpperBound[ordinal()][lat.ordinal()];
	}
	
	/**
	 * Returns the greatest lower bound of two lattice points.
	 */
	public EBooleanLattice greatestLowerBound(EBooleanLattice lat) {
		return GreatestLowerBound[ordinal()][lat.ordinal()];
	}
	
	/**
	 * Returns the possible results of a logical AND between two booleans.
	 * BOTTOM will be returned if either operand is BOTTOM.
	 */
	public EBooleanLattice and(EBooleanLattice lat) {
		return And[ordinal()][lat.ordinal()];
	}
	
	/**
	 * Returns the possible results of a logical AND between two booleans.
	 * BOTTOM will be returned if either operand is BOTTOM.
	 */
	public EBooleanLattice or(EBooleanLattice lat) {
		return Or[ordinal()][lat.ordinal()];
	}
	
	/** Returns true if this is YES or BOTTOM */
	public boolean definitely() {
		switch (this) {
		case YES:
		case BOTTOM:
			return true;
		default:
			return false;
		}
	}
	/** Returns true if this is NO or BOTTOM */
	public boolean definitelyNot() {
		switch (this) {
		case NO:
		case BOTTOM:
			return true;
		default:
			return false;
		}
	}
	/** Returns true if this is YES or MAYBE */
	public boolean maybe() {
		switch (this) {
		case YES:
		case MAYBE:
			return true;
		default:
			return false;
		}
	}
	/** Returns true if this is NO or MAYBE */
	public boolean maybeNot() {
		switch (this) {
		case NO:
		case MAYBE:
			return true;
		default:
			return false;
		}
	}
}
