package dk.brics.xact.compiler;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Compiler from XACT to Java.
 * Transforms files with extension <tt>.xact</tt> into ordinary Java code (<tt>.java</tt>).
 */
public class Main {
	
	private Main() {}
	
	/**
	 * Compiles XACT syntax into Java.
	 */
	public static void main(String[] args) throws IOException {
		try {
			main2(args);
		} catch (SetupException ex) {
			System.err.println(ex.getMessage());
			System.exit(1);
		}
	}
	
	/**
	 * Compiles XACT syntax into Java.
	 * @throws SetupException if invalid arguments are provided
	 */
	public static void main2(String[] args) throws IOException, SetupException {
		if (args.length == 0) {
			System.out.println(
					"Usage: java dk.brics.xact.compiler.Main [ARGS]\n" +
					"\n" +
					"Arguments:\n" +
					"  -cp=<classpath>     Append this to classpath.\n" +
					"  -dir=<path>         Compile all .xact files in this directory,\n" +
					"                      including all subdirectories.\n" +
					"                      Multiple -dir arguments may be given.\n" +
					"  -out=<path>         Put the output in this directory.\n" +
					"  <filename>          Compile this .xact file.\n"
					);
			System.exit(0); // FIXME: clean up use of SetupException and System.exit in main methods (also in dk.brics.xact.analysis.Main)
		}
		
		String[] cp = System.getProperty("java.class.path").split(File.pathSeparator);
		List<String> classpaths = new ArrayList<String>(Arrays.asList(cp));
		List<File> files = new ArrayList<File>();
		
		String outputdir = null;
		
		for (String arg : args) {
			if (arg.startsWith("-dir=")) {
				String dir = arg.substring("-dir=".length());
				files.addAll(FileUtil.findAllFilesWithExtensions(new File(dir), true, ".xact", ".java"));
			}
			else if (arg.startsWith("-cp=")) {
				String cparg = arg.substring("-cp=".length());
				classpaths.add(cparg);
			}
			else if (arg.startsWith("-out=")) {
				String path = arg.substring("-out=".length());
				if (outputdir != null)
					throw new SetupException("At most one -out argument may be given");
				outputdir = path;
			}
			else if (arg.endsWith(".xact") || arg.endsWith(".java")) {
				files.add(new File(arg));
			}
			else {
				throw new SetupException("Unrecognized argument " + arg);
			}
		}
		
		if (outputdir == null) {
			outputdir = ".";
		}
		
		List<String> compilerArgs = new ArrayList<String>();
		String classpath = implode(classpaths, File.pathSeparator);
		compilerArgs.add("-cp");
		compilerArgs.add(classpath);
		compilerArgs.add("-prettyprint");
		compilerArgs.add(outputdir);
		compilerArgs.add("-xact");
		for (File file : files) {
			compilerArgs.add(file.getPath());
		}
		
		dk.brics.java5.compiler.Main.main(compilerArgs.toArray(new String[0]));
	}	
	
	private static String implode(Iterable<?> it, String delim) {
		boolean first=true;
		StringBuilder b = new StringBuilder();
		for (Object s : it) {
			if (!first)
				b.append(delim);
			else
				first = false;
			b.append(s);
		}
		return b.toString();
	}
}
