package dk.brics.xact.operations;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import dk.brics.xact.Element;
import dk.brics.xact.XML;

/**
 * Indents XHTML output for the purpose of making it readible as possible, while attempting
 * not to change its rendering (see note below) using the following rules:
 * <ul>
 * <li>Anything not in the XHTML namespace will be indented.
 * <li>Inline elements such as <tt>&lt;a&gt</tt> and <tt>&lt;span&gt;</tt> will not be indented.
 * <li>The elements <tt>&lt;pre&gt;</tt>, <tt>&lt;textarea&gt;</tt>, and <tt>&lt;code&gt;</tt> will
 *     be considered whitespace-sensitive.
 * <li>Elements with only simple content will not be indented. For example, here are two indentations
 *      of a <tt>&lt;title&gt;</tt> element:
 *      <pre>
 *          &lt;title&gt;Hello world&lt;/title&gt;
 *          &lt;title&gt;
 *              Hello 
 *              world
 *          &lt;/title&gt;
 *      </pre>
 *    The linebreak in the title causes it to become indented.
 * </ul>
 * The indentation uses four whitespaces per level and the system
 * property <tt>line.separator</tt> as the linebreaker.
 * <p/>
 * Please note that CSS can cause other XHTML elements to become whitespace-sensitive, so this
 * is not generally a sound indentation for XHTML (ie. it can change the layout of the page).
 * Therefore, this is mainly intended for debugging.
 */
public class XHTMLIndentation implements XMLIndentation {
    
    private static final String XHTML_NAMESPACE = "http://www.w3.org/1999/xhtml";
    
    private Set<String> dontIndent = new HashSet<String>(Arrays.asList(
            "a","abbr","acronym","b","bdo","big","br","cite","code",
            "del","dfn","em","i","img","ins","label","map","kbd","object",
            "q","ruby","samp","small","span","strong","sub","sup","textarea",
            "tt","var",
            "u" // <u> is deprecated, but in case someone uses it anyway, don't indent it
    ));
    private Set<String> sensitive = new HashSet<String>(Arrays.asList(
            "pre","textarea","code"
    ));
    
    private boolean wantIndent(Element e) {
        return !e.getNamespace().equals(XHTML_NAMESPACE) || !dontIndent.contains(e.getLocalName());
    }

    private boolean hasBlockContent(Element n) {
        for (XML node = n.getFirstChild(); node!=null; node = node.getNextSibling()) {
            if (node.isElement() && wantIndent(node.asElement())) {
                return true;
            }
            if (node.isText()) {
                String s = node.asText().getString();
                if (s.indexOf('\n') != -1 || s.indexOf('\r') != -1)
                    return true;
            }
        }
        return false;
    }
    
    public boolean isWhitespaceSensitive(Element element) {
        return element.getNamespace().equals(XHTML_NAMESPACE) && sensitive.contains(element.getLocalName());
    }
    
    public IndentType getIndentType(Element element) {
    	if (wantIndent(element)) {
    		if (hasBlockContent(element))
    			return IndentType.BLOCK;
    		else
    			return IndentType.INLINE_BLOCK;
    	} else {
    		return IndentType.NONE;
    	}
    }
    
    public String getLineBreak() {
        return System.getProperty("line.separator");
    }
    
    public String getIndentation() {
        return "    ";
    }
}
