package dk.brics.xact.operations;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.util.Iterator;
import java.util.Map;

import dk.brics.xact.*;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

import dk.brics.misc.Origin;

/**
 * Parser for XML templates.
 */
public class XMLParser { // TODO: improve template syntax error messages?

	/**
	 * Namespace for XACT gap elements.
	 */
    public static final String XACT_NAMESPACE = "http://www.brics.dk/Xact";

    private static EntityResolver resolver;

    private XMLParser() { }

    /**
     * Sets the entity resolver.
     */
    public static void setResolver(EntityResolver resolver) {
        XMLParser.resolver = resolver;
    }

    /**
     * Parses the given XML template or document from its textual representation.
     * If <code>template</code> is set, the global and thread local namespace declarations apply, and gaps are recognized.
     * If <code>namespaces</code> is non-null, this map overrides the global and thread local namespace declarations.
     *
     * @throws IOException          if an I/O error occurs
     * @throws XMLTemplateException if the string is not a syntactically wellformed XML template or document
     */
    public static XML parse(InputStream in, String encoding, boolean template, Map<String, String> namespaces, Origin origin)
            throws IOException {
        String s = wrapAndConvertGaps(in, encoding, template, namespaces, origin);
        ParseHandler h = new ParseHandler(origin);
        try {
            XMLReader r = XMLReaderFactory.createXMLReader();
            r.setContentHandler(h);
            r.setErrorHandler(h);
            r.setProperty("http://xml.org/sax/properties/lexical-handler", h);
            if (resolver != null) {
                r.setEntityResolver(resolver);
            } else {
                //The standard resolver fetches the XHTML from the W3C server. This fails at random since
                //W3C no longer want people to do that. Therefore, ignore the DTD altogether.
                r.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd", false);
                r.setFeature("http://xml.org/sax/features/validation", false);
            }
            r.parse(new InputSource(new StringReader(s)));
            if (template) {
                XML x = h.getRoot();
                if (x instanceof Element)
                    x = ((Element) x).getFirstChild();
                if (x == null) {
                    x = new Text("");
                }
                return x;
            } else {
                return h.getRoot();
            }
        } catch (SAXException e) {
            throw new XMLTemplateException(e, origin);
        }
    }

    private static String wrapAndConvertGaps(InputStream in, String encoding, boolean template, Map<String, String> namespaces, Origin origin)
            throws XMLException, IOException {
        StringBuilder b = new StringBuilder();
        final Reader r = new BufferedReader(new InputStreamReader(in, encoding));
        Iterator<Character> i = new Iterator<Character>() {

            int c = r.read();

            public boolean hasNext() {
                return c != -1;
            }

            public Character next() {
                char d = (char) c;
                try {
                    c = r.read();
                } catch (IOException e) {
                    throw new XMLException(e);
                }
                return d;
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
        if (template) {
            b.append("<XML");
            if (namespaces != null) {
                for (Map.Entry<String, String> me : namespaces.entrySet()) {
                    makeNamespaceDecl(b, me.getKey(), me.getValue());
                }
            } else {
                for (Map.Entry<String, String> me : XML.getThreadNamespaceMap().entrySet()) {
                    makeNamespaceDecl(b, me.getKey(), me.getValue());
                }
                for (Map.Entry<String, String> me : XML.getNamespaceMap().entrySet()) {
                    if (!XML.getThreadNamespaceMap().containsKey(me.getKey())) {
                        makeNamespaceDecl(b, me.getKey(), me.getValue());
                    }
                }
            }
            b.append(">");
        }
        try {
            while (i.hasNext()) {
                char c = i.next();
                if (c == '<' && i.hasNext()) {
                    c = i.next();
                    if (c == '?') { // processing instruction
                        b.append("<?");
                        if (i.hasNext()) {
                            char c1 = i.next();
                            b.append(c1);
                            while (i.hasNext()) {
                                char c2 = i.next();
                                b.append(c2);
                                if (c1 == '?' && c2 == '>') {
                                    break;
                                }
                                c1 = c2;
                            }
                        }
                    } else if (c == '!') { // comment
                        b.append("<!");
                        if (i.hasNext()) {
                            char c1 = i.next();
                            b.append(c1);
                            if (i.hasNext()) {
                                char c2 = i.next();
                                b.append(c2);
                                while (i.hasNext()) {
                                    char c3 = i.next();
                                    b.append(c3);
                                    if (c1 == '-' && c2 == '-' && c3 == '>') {
                                        break;
                                    }
                                    c1 = c2;
                                    c2 = c3;
                                }
                            }
                        }
                    } else if (template && c == '[') { // template gap
                        makeGap("tgap", b, i, "gap=", namespaces, origin);
                        if (i.hasNext()) {
                            i.next();
                        }
                    } else { // tag starting
                        b.append('<');
                        b.append(c);
                        while (i.hasNext()) {
                            c = i.next();
                            if (c == '>') {
                                b.append(c);
                                break;
                            } else if (isWS(c)) { // attributes starting
                                StringBuilder attrgaps = new StringBuilder();
                                StringBuilder buf = new StringBuilder();
                                boolean inattrname = false;
                                while (i.hasNext()) {
                                    c = i.next();
                                    if (c == '"') { // attribute value starting, enclosed by "
                                        inattrname = false;
                                        b.append(' ');
                                        if (buf.length() > 0) {
                                            b.append(buf);
                                            buf = new StringBuilder();
                                        }
                                        b.append(c);
                                        while (i.hasNext()) {
                                            c = i.next();
                                            b.append(c);
                                            if (c == '"') {
                                                break;
                                            }
                                        }
                                    } else if (c == '\'') { // attribute value starting, enclosed by '
                                        inattrname = false;
                                        b.append(' ');
                                        if (buf.length() > 0) {
                                            b.append(buf);
                                            buf = new StringBuilder();
                                        }
                                        b.append(c);
                                        while (i.hasNext()) {
                                            c = i.next();
                                            b.append(c);
                                            if (c == '\'') {
                                                break;
                                            }
                                        }
                                    } else if (template && c == '[') { // attribute gap
                                        inattrname = false;
                                        makeGap("agap", attrgaps, i, buf.toString(), namespaces, origin);
                                        buf = new StringBuilder();
                                    } else if (c == '>') { // tag ending
                                        if (inattrname && !buf.toString().trim().equals("/")) {
                                            throw new XMLTemplateException("malformed XML, expected an attribute value");
                                        }
                                        b.append(buf).append(c).append(attrgaps);
                                        attrgaps = new StringBuilder();
                                        break;
//                                    } else if (isWS(c)) {
//                                        if (inattrname) {
//                                            throw new XMLTemplateException("malformed XML, unexpected whitespace");
//                                        }
//                                        if (buf.length() > 0) {
//                                            b.append(buf);
//                                            buf = new StringBuilder();
//                                        }
                                    } else {
                                    	if (!isWS(c)) {
                                    		inattrname = true;
                                    	}
                                        buf.append(c);
                                    }
                                }
                                break;
                            }
                            b.append(c);
                        }
                    }
                } else {
                    b.append(c);
                }
            }
            if (template) {
                b.append("</XML>");
            }
        } finally {
            r.close();
        }
        return b.toString();
    }

    private static void makeNamespaceDecl(StringBuilder b, String prefix, String ns) {
        b.append(" xmlns");
        if (prefix.length() > 0) {
            b.append(':').append(prefix);
        }
        b.append("=\"").append(ns).append("\"");
    }

    private static void makeGap(String name, StringBuilder b, Iterator<Character> i, String g, Map<String,String> namespaces, Origin origin) {
        char c;
        StringBuilder typegap = new StringBuilder();
        while (i.hasNext()) {
            c = i.next();
            if (c == ']') {
                break;
            }
            typegap.append(c);
        }
        String tg = typegap.toString().trim();
        String type;
        String gap;
        int j = tg.indexOf(' ');
        if (j == -1) {
            j = tg.indexOf('\t');
        }
        if (j == -1) {
            j = tg.indexOf('\r');
        }
        if (j == -1) {
            j = tg.indexOf('\n');
        }
        if (j != -1) {
            type = tg.substring(0, j).trim();
            gap = tg.substring(j).trim();
        } else {
            type = null;
            gap = tg;
        }
        b.append("<xact:").append(name).append(' ');
        if (type != null) {
            b.append("xact:type=\"").append(expandQName(type, namespaces, origin)).append("\" ");
        }
        b.append(g).append("\"").append(gap).append("\" xmlns:xact=\"").append(XACT_NAMESPACE).append("\"/>");
    }
    
    private static String expandQName(String type, Map<String,String> namespaces, Origin origin) {
        if (namespaces == null) {
            return XMLValidator.expandQName(type, XML.getThreadNamespaceMap(), XML.getNamespaceMap(), origin);
        } else {
            return XMLValidator.expandQName(type, namespaces, null, origin);
        }
    }
    
    private static boolean isWS(char c) {
        return " \t\r\n".indexOf(c) != -1;
    }
}
